/***********************************
 *                                 
 *  pfilter.c                      
 * 
 *  pfilter.c defines the code for a filter program designed
 *  to work for the phaser 560 printer from Tektronix.  This
 *  filter is designed to work with LPRng as created and 
 *  maintained by Patrick Powell. Tested with
 *  version 3.5.1
 *
 *  Written by Steve Wilson 8-12-98
 *  Written on SunOS 5.6
 *
 */
/***********************************
 * 
 *  Version: 1.0.2
 *
 */
/***********************************
 *
 *  DESIGN GOALS: 
 * 
 *  I designed this filter to use posix threads to speed
 *  up the read/write portion of the filter.  The 
 *  basic design of the filter follows this outline
 *  to print to the phaser560:
 *    1.  Open a TCP socket to the printer, send a pagerequest in postscript, close
 *        the TCP socket.
 *    2.  Open a second TCP socket to the printer, send the print job, close the
 *        TCP socket.
 *    3.  Open a third TCP socket to the printer, send the final pagerequest in 
 *        postscript, close the TCP socket.
 *  In addition, I have modified my code so that the close(socket) call instead use
 *  shutdown(socket) followed by read until bytes read == 0. (PLEASE SEE 
 *  NOTES.TXT)
 *  
 *  The filter has three main threads of execution.  1. The first thread
 *  handles submission and interpretation of page requests to the printer.
 *  2. The second thread handles reading of input data from LPRng. 3.  The 
 *  third thread handles writing the printjob data to the printer.
 *
 *  This filter maintains several global structures/variables.  The most 
 *  important is the read/write buffer.  This buffer is divided into
 *  slots or buckets.  The second structure is a control structure that 
 *  allows access into the read/write buffer via the slots.  The idea
 *  behind the read/write buffer is for the read thread to fill the buffer
 *  while the write thread unloads the buffer to the printer.  Each
 *  thread fills or writes from a bucket at a time.  The buffer control
 *  structure keeps track of whether or not a given bucket has data
 *  in it.  IE, For a given bucket, when the read thread examines it,
 *  the thread needs to know whether the bucket can accept new data
 *  or not.  Similarly, the write thread needs to know whether a
 *  given bucket has valid data that needs to go to the printer.
 *
 *  The filter follows this basic flow of execution:
 *    1.  Parse the command line parameters from LPRng and initialize all 
 *        necessary variables.
 *    2.  Open the log accounting file
 *    3.  Create the page request thread
 *    4.  Create the read thread
 *    5.  Create the write thread
 *    6.  Wait for successful termination of the page request thread
 *    7.  Write final accounting and log messages
 *    8.  Clean up
 */
/***********************************
 *
 *  ISSUES YET TO BE RESOLVED: (AND ISSUES FIXED/CHANGED STARTING 11-24-98)
 * 
 *  - More error checking ! - this code only does a slight
 *    amount of error checking and certainly could check more
 *    proactively.
 *
 *  - Support for transparency queue - At the Beckman Institute
 *    we maintain "transparency queues" for our Unix users.
 *    We do this by sending a tray change command in postscript
 *    before sending the user's job.  I need to incorporate support
 *    for this feature yet.
 *
 *    - Resolved 11-24-98
 *      Implemented support for tray switching commands.  These include
 *      forcing the printer to print from a specified tray, as well
 *      as requesting prints to come from a particular type of
 *      media, eg. transparency. 
 *  
 *  - More robust support for multiple printers.  At this time
 *    the filter uses the -P command line argument from
 *    LPRng to create the printer hostname.  (We have lp=/dev/null
 *    set in printcap).  Additionally the filter parses off
 *    the -P option and appends .beckman.uiuc.edu to the
 *    name.  While this works for us it doesn't seem to fit
 *    cleanly with LPRng.
 * 
 *    - Partially Resolved 11-24-98
 *      Changed -P parse algorithm to append -P value to #define DOMAIN
 *      variable.  Thus, you can easily change the domain for a printer
 *      at compile time.
 *      Other prior problems still apply.
 *
 *  - Printer Power Cycles - As is, this code will successfully
 *    recover from a power cycle of the printer.  It will wait
 *    until the printer comes online again and query for the 
 *    final page count.  This all happens successfully.  
 *    Unfortunately, due to some other error not quite yet
 *    understood, the filter later aborts with a "BUS ERROR"
 *    when certain system calls such as localtime_r(), or 
 *    perror() are called.  Let me emphasize, this does not
 *    affect the normal operation of the filter.  This situation
 *    only occures when the printer has been power cycled.  
 *    It appears that the write() call that eventually aborts
 *    is leaving something in an unstable state.  If you know 
 *    about bus errors, or how sockets behave when the remote
 *    host has forcefully terminated the connection(ie powercycle)
 *    please contact me to help resolve this issue.
 *
 *    - Resolved 11-24-98
 *      I found that the printer no longer causes a "BUS ERROR" in
 *      the filter once extraneous protocols are disabled on the printer.
 *      We do not use any protocol other than TCP/IP.  Disabling all others
 *      caused this bug to go away.  Thus, the filter will catch
 *      the pagecount in the event of a powercycle once the machine
 *      has come back up.
 *
 *  - We know of several other issues related to LPRng behavior that
 *    could potentially become a problem with this filter as it exists
 *    now.  Most of this problems stems from our own incomplete
 *    knowledge of how LPRng behaves in certain circumstances.  Again
 *    these issues are being addressed.  If you would like to know
 *    more specifically what we are investigating contact me.  Or,
 *    if you would like to report a bug, or bring something else to
 *    my attention, please contact me.
 *
 *  - Debugging - For the time being the code can be compiled in
 *    one of two modes, standard mode or DEBUG mode.  To compile for
 *    debug mode use the -DDBG option when compiling this file.
 *    A more sophisticated debugging method will appear in future
 *    versions of this filter.
 *
 *  - FIXED 11-24-98
 *    Filter did not check for printer status at start up and aborted
 *    if printer wasn't availalbe.  Changed to use a progressive timeout
 *    structure.  Filter queries UDP port till printer reports status idle.
 *    Then, the filter starts the TCP series of queries.  Also, I have 
 *    changed the pagecount thread to loop on each pagecount request until
 *    the printer reports back a pagecount.  This prevents the filter
 *    from aborting in certain error conditions.
 *
 *  - Added -Z and -T options.  Filter now accepts the following -T 
 *    options; -Tbanner -> prints a banner page after the job, 
 *    -Tbanner=last -> prints a banner page after the job, -Tbanner=first
 *    -> prints a bannerpage before the job, -Tbanner=off -> does not
 *    print a banner, -Tbanner=on -> prints a bannerpage after the 
 *    job. With no -Tbanner option, the filter does not print a banner
 *    page.  Banner pages can be specified in as a postscript file 
 *    as indicated from the :bp= line in printcap.  The filter looks 
 *    for the following tags in the banner postcript; **PNAME** ->
 *    printer name, **TYPE** -> printer type(this is set to Tektronix
 *    Phaser560), **UNAME** -> name of user submiting job, **HNAME** ->
 *    name of host job submitted from, **JNAME** -> name of data file
 *    print job spooled from, **DT** -> date and time.  If a banner
 *    page is specified the accounting routine automatically subtracts
 *    one from the number of pages printed.
 *
 *    The filter also supports the following -Z options; -Ztransparency
 *    -> forces the printer to draw paper from available transparency 
 *    trays, -Zglossy -> forces the printer to draw paper from a 
 *    predermined glossy paper tray.  The glossy paper tray can be
 *    set as an option down in the #defines. -Ztransparency and -Zglossy
 *    are mutually exclusive options.  The filter will not function
 *    appropriately with both specified.
 *
 *  - We have done minor testing of the filter with the Tektronix 
 *    Phaser600 model of printers and seems to function well.  However, 
 *    be advised that all tray switching commands(includes -T and 
 *    -Z options) shouldn't be used with the Phaser600.
 *    By far and above, this filter is only designed to work with the
 *    Phaser560 model of printers.  If you have luck with other models,
 *    great.
 **********************************************************************/


#include<sys/socket.h>
#include<netinet/in.h>
#include<stdlib.h>
#include<sys/types.h>
#include<netdb.h>
#include<errno.h>
#include<arpa/inet.h>
#include<unistd.h>                                
#include<stdio.h>
#include<string.h>
#include<pthread.h>
#include<fcntl.h>
#include<time.h>
#include"tray.h"



/*  Global defines.  
 *    bzero - bzero is a function that doesn't seem to exist on SunOS 5.6 in a include file
 *            so I have defined it here.
 *    PAGECOUNT - postscript command string to query printer for pagecount
 *    IOBUFFSZ - size of bucket
 *    PJOBSZ - size of the read/write buffer
 *    EBUFFSZ - size of the printer return buffer
 *    TCPPORT - TCP to connect with on printer
 *    UDPPORT - UDP port to connect with on printer
 *    CHUNK - number of buckets in read/write buffer, this could be calculated in the program, but I
 *            found it easier to set explicitly here for debugging purposes.
 *    FIRST - used to flag whether or not the banner should be printed before the job.
 *    LAST - used to flag whether or not the banner should be printed after the job.
 *    TRANSTRAY - used to indicate which tray should be used for the transparency tray
 *    BANNERTRAY - used to indicate which tray to print the banner on.
 *    GLOSSYTRAY - used to indicate which tray to print glossy paper on.
 *    ANYTRAY - used to switch printer back to default paper tray.
 *    DOMAIN - used to specify domain to append printer name to 
 *
 *    trays use the following scheme as defined in tray.h.  this list is *NOT* authoritative
 *    and is only provided as a convenience.  please see tray.h.
 *        default tray - 1, lower tray - 2, middle tray - 3, upper tray - 4,
 *        transparency tray - 5
 *    
 *    The idea behind the buffer is; 1 PJOBSZ buffer with CHUNK buckets of IOBUFFSZ bytes / bucket
 */
#ifndef HAVE_BZERO
#define bzero(ptr, n) memset(ptr, 0, n)
#endif
/*
#define PAGECOUNT "%!PS\nfalse 0 startjob pop\n(%%[ pagecount: )print statusdict/pagecount get exec(               )cvs\nprint\n(, 1 1 ]%%)= flush\n"
*/
#define PAGECOUNT "false 0 startjob pop\n(%%[ pagecount: )print statusdict/pagecount get exec(                )cvs\nprint\n(, 1 1 ]%%)= flush\n(***)= flush\n"
#define STATUS "\r\n"
#define IOBUFFSZ 0x1400
#define PJOBSZ 0x100400
#define RBUFFSZ 0xC8
#define TCPPORT 9100
#define UDPPORT 9101
#define CHUNK 204 
#define FIRST 0
#define LAST 1
#define TRANSTRAY 5
#define BANNERTRAY 2
#define GLOSSYTRAY 2
#define ANYTRAY 1
#define DOMAIN ""
/* #define DOMAIN ".beckman.uiuc.edu" */


/*  GLOBAL VARIABLES */

/*  boolean data type */
typedef enum bl {f, t} bl; 


/*  done[] - when each thread terminates successfully they mark their done
            variable as true.
    abrt[] - when each thread terminates unsuccessfully they mark abrt as true
    banner - t/f depending on whether the filter should generate a banner page.
    trans - t/f depending on whether the filter should print on transparency
    glossy - t/f depending on whether the filter should print on glossy paper
*/

bl done[3], abrt[3], banner, trans, glossy;

/*  buffer control structure 
 *    wash - wash is set to true if the bucket has valid data in it.  The idea is
 *           that the once the bucket has valid data in it, the write thread
 *           must write that data, ie wash the bucket.  If wash is false that
 *           means that write should not wash the bucket and the read thread
 *           can fill the bucket.
 *    size - this contains the number of valid data bytes in the bucket.  The
 *           read system call does not necessarily read the number of bytes
 *           requested.  I call read(buffer, IOBUFFSZ), and store in size
 *           the number of bytes actually read.
 */
struct ctl{
    bl wash;
    unsigned int size;
} ctl[CHUNK + 1];

/*  SOCKET VARIABLES */
struct sockaddr_in praddr, udppraddr;
struct in_addr ipaddr;
struct hostent *prent;
int prsocket;

/*  COUNTERS/VALUE HOLDERS 
 *    p1 - holds the page count from the first page request to the printer.
 *    p2 - holds the second page count from the second page request to the 
 *         printer.
 *    jnumber - holds the jobnumber as passed in on the command line from 
 *              LPRng
 *    end - holds the end of the job that the banner should be printed on,
 *          values are FIRST/LAST
 */

int p1, p2, jnumber, end;


/*  BUFFERS 
 *    pjob - read/write buffer
 *    pname - hostname of printer, passed in from LPRng and modified 
 *            slightly.
 *    acctfilename - hold the name of the accounting file as passed
 *                   from LPRng
 *    jfilename - holds the path/name of the data file as passed
 *                from LPRng
 *    usname - holds the user name for the current print job as
 *             passed from LPRng
 *    hname - holds the name of the host usname submitted the
 *            job from as passed from LPRng
 *    bannername - holds the name of file to read the banner page 
 *                 from
 */
char *pjob;
char *pname, *acctfilename, *jfilename, *usname, *hname, *bannername;

/*  DATE RELATED 
 *    mnth - used in calls that write to log files to stamp time/date
 */
char *mnth[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", 
		 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

/*  TIME RELATED 
 *    t1 - holds the time stamp returned from gettimeofday,
 *         Due to the localtime_r/bus error issues this is
 *         used to delay account file logging until the end
 *         of the filter.
 *    t2 - holds the second time stamp returned from
 *         gettimeofday
 *
 *    These two are stamped after a pagerequest call to the
 *    printer.
 */
struct timeval t1, t2;


/*  MUTEX VARIABLES 
 *    skt - protects all access to the socket variables
 *    wsh - protects all access to the buffer control
 *          structure
 *    preq - prevents the write thread from running until
 *           first page request completes
 */
pthread_mutex_t skt = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t wsh = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t preq = PTHREAD_MUTEX_INITIALIZER;

/*  FUNCTION PROTOTYPES */
int begin(int, char **, char **);
char *con(int, char*);
int account(void);
int pstatus(int);
void *pagereq(void *);
void *readinp(void *);
void *writeop(void *);
int genbanner(int, int);
void insrt(char *, char *, int, int *, unsigned long *);
int tselect(int,int);
int a(int *);
int cls();

/* MAIN FUNCTION 
 *   Main gets the arguments from the commandline and sends them to begin to process. Begin will process
 *   the arguments and initialze the various data structures for the filter.  Next main will create three
 *   threads.  Thread 1 will handle getting page counts from the printer.  Thread 2 will handle reading
 *   the input file into the pjob buffer.  Thread 3 will write the data in the pjob buffer to the printer.
 *   Main also opens the accounting file writes the page accounting.  It also writes a start and an end
 *   statement to the log file.  The accounting code should be moved into thread 1, and will be as soon
 *   as the localtime/write()/bus error issue is resolved.
 */
/* Modified 9-17-98
 *   Moved the accounting logging code into a separate routine account.
 */

int main(int argc, char **argv, char **env)
{
  pthread_t preq, rinp, wrop;
  int bres, pres, ires, i, j;
  time_t rtime;
  struct tm ltime, *hold;
  char *elist[] = 
  {
     "Error: unable to open accounting file",
  };
  char *buff[4];
  

#ifdef DBG
  fprintf(stderr, "Calling begin\n");
#endif

  for(i = 0; i < 4; i++){
    buff[i] = (char *)calloc(4, sizeof(char));
  }
  

/*  call begin to parse the arguments and initialize the data structures */
  if((bres = begin(argc, argv, env)) < 0){
    fprintf(stderr, "Error %d initializing filter\n", bres);
    exit(-1);
  }
#ifdef DBG
  fprintf(stderr, "After Begin\n");
#endif

/*  get the time for first log file write */
  rtime = time(NULL);
  localtime_r(&rtime, &ltime);

/*  write the first log message */
  /*  fprintf(stderr, "\nSTART-test: job number %d (dfile: %s) for %s@%s on %s at %s %d %d:%d:%d\n", jnumber, jfilename, usname, hname, pname, mnth[ltime.tm_mon], ltime.tm_mday, ltime.tm_hour, ltime.tm_min, ltime.tm_sec); 
   */
  fprintf(stderr, "\nSTART: job number %d (dfile: %s) for %s@%s on %s at %s %s %s:%s:%s\n", jnumber, jfilename, usname, hname, pname, mnth[ltime.tm_mon], con(ltime.tm_mday, buff[0]), con(ltime.tm_hour, buff[1]), con(ltime.tm_min, buff[2]), con(ltime.tm_sec, buff[3])); 


/*  check to see if printer is at status idle.
 *  sleep(1) first to clear all socket info on printer.
 */ 
  ires = pstatus(5);
  i = 0;
  while(ires < 0){
    if(i < 5){
      fprintf(stderr, "Waiting: Printer not at status idle, sleeping for 60 seconds\n");
      sleep(60);
    }
    else{
      fprintf(stderr, "Waiting:  Printer not at status idle, sleeping for 300 seconds\n");    
      sleep(300);
    }
    i++;
    if(i < 0) i = 5; 
    ires = pstatus(5);   
  }
  

/*  create the pagerequest thread */
  if((pres = pthread_create(&preq, NULL, pagereq, NULL)) > 0){
    fprintf(stderr, "Error %d starting thread preq\n", pres);
    exit(-1);
  }

/*  create the read thread */
  if((pres = pthread_create(&rinp, NULL, readinp, NULL)) > 0){
    fprintf(stderr, "Error %d starting thread rinp\n", pres);
    exit(-1);
  }

/*  create the write thread */
  if((pres = pthread_create(&wrop, NULL, writeop, NULL)) > 0){
    fprintf(stderr, "Error %d starting thread wrop\n", pres);
    exit(-1);
  }

/*  wait for the page request thread to terminate */
  pthread_join(preq, NULL);
 
/*  testing account function */
  if(account() < 0){
    fprintf(stderr, "Error running page accounting\n");
    exit(-1);
  }
  
/*  get time for last write to log file */
  rtime = time(NULL);
  localtime_r(&rtime, &ltime);

#ifdef DBG
  fprintf(stderr, "Returned from end of job time calls\n");
  fprintf(stderr, "Writing end of job\n");
#endif

/*  write last entry to log file */
  /*
  fprintf(stderr, "END-test: job number %d, (dfile: %s) for %s@%s on %s at %s %d %d:%d:%d\n", jnumber, jfilename, usname, hname, pname, mnth[ltime.tm_mon], ltime.tm_mday, ltime.tm_hour, ltime.tm_min, ltime.tm_sec); 
  */

  fprintf(stderr, "END: job number %d, (dfile: %s) for %s@%s on %s at %s %s %s:%s:%s\n", jnumber, jfilename, usname, hname, pname, mnth[ltime.tm_mon], con(ltime.tm_mday, buff[0]), con(ltime.tm_hour, buff[1]), con(ltime.tm_min, buff[2]), con(ltime.tm_sec, buff[3])); 


/*  call the cleanup routine */
  cls();
  sleep(2); 
  exit(0);
}

/* BEGIN
 *   Begin accepts int argc, and int argv from the commandline.  Begin parses for the following flags: -e for the
 *   job filename on disk, -a for the accounting file name, -j for the job number, -n for the user name, -h for 
 *   name of the host the user submitted the job from, -P for the printer name.  The filter supports the following
 *   -T options: -Tbanner - generates banner page at the end of job, -Tbanner=on - generates banner page
 *   at the end of job, -Tbanner=first - generates banner page at the beginning of a job, -Tbanner=last -
 *   generates a banner page at the end of a job, -Tbanner=off - turns the banner off. By default, with
 *   no -T option, a banner will not be printed.  The filter supports the following -Z options; -Ztransparency -
 *   puts the printer into mode where it will generate a job from the transparency tray, -Zglossy - puts the
 *   printer into a mode where it will generate a job from the glossy paper tray.  These options are
 *   mutually exclusive.  You cannot pass -Zglossy -Ztransparency.  
 *   Our queue names defined in printcap are the first name of the printer's host name.
 *   eg, in printcap phaser560|phaser560t| ... -> phaser560.beckman.uiuc.edu. In addition, 
 *   begin initializes the other global variables needed by the filter.
 */
int begin(int argc, char **argv, char **env)
{
  int i = 1, j = 0;
  char *s, *t1, *t2, *t3, *t4;
  char *elist[] = 
  {
    "Error: Unable to open accounting file"
  };


#ifdef DBG
  fprintf(stderr, "\nStarting new FILTER CYCLE\n");
  fprintf(stderr, "\nargc = %d\n", argc);
#endif

/*  banners are turned off by default 
 *  transparency/glossy also turned off by default
 *  the banner will default to printed last.
 */
  banner = f;
  trans = f;
  glossy = f;
  end = LAST;


  for(i = 1; i < argc; i++){
    s = argv[i];
#ifdef DBG
    fprintf(stderr, "arvg[%d] = %s\n", i, s);
#endif
    if(*s == '-'){
      switch(*(s + 1))
      {
        case 'e':
	  jfilename = (char *)calloc(strlen(argv[i]) - 1, sizeof(char));
	  memcpy(jfilename, s + 2, strlen(argv[i]) - 2);
#ifdef DBG
	  fprintf(stderr, "jfilename: %s\n", jfilename);
#endif
	  break;
        case 'a':
	  acctfilename = (char *)calloc(strlen(argv[i]) - 1, sizeof(char));
	  memcpy(acctfilename, s + 2, strlen(argv[i]) - 2);
#ifdef DBG
 	  fprintf(stderr, "acctfilename: %s\n", acctfilename);
#endif
	  break;
        case 'j':
          jnumber = atoi(argv[i] + 2);
          break;
        case 'n':
 	  usname = (char *)calloc(strlen(argv[i]) - 1, sizeof(char));
     	  memcpy(usname, s + 2, strlen(argv[i]) - 2);
	  break;
        case 'h':
	  hname = (char *)calloc(strlen(argv[i]) - 1, sizeof(char));
	  memcpy(hname, s + 2, strlen(argv[i]) - 2);
          break;
        case 'P':
	  pname = (char *)calloc(strlen(argv[i]) + strlen(DOMAIN) - 1, sizeof(char));
	  memcpy(pname, s + 2, strlen(argv[i]) - 2);
	  memcpy(pname + strlen(argv[i]) - 2, DOMAIN, strlen(DOMAIN));
#ifdef DBG
          fprintf(stderr, "pname = %s\n", pname);
#endif
	  break;
        case 'T':
	  if((t4 = strstr(s + 2, "banner=first")) != NULL){
#ifdef DBG
	    fprintf(stderr, "s + 2 = %s, should be banner=first\n", s + 2);
#endif
	    banner = t;
	    end = FIRST;
	  }
	  else if((t4 = strstr(s + 2, "banner=last")) != NULL){
#ifdef DBG
	    fprintf(stderr, "s + 2 = %s, should be banner=last\n", s + 2);
#endif
	    banner = t;
	    end = LAST;
	  }
	  else if((t4 = strstr(s + 2, "banner=on")) != NULL){
#ifdef DBG
	    fprintf(stderr, "s + 2 = %s, should be banner=on\n", s + 2);
#endif
	    banner = t;
	    end = LAST;
	  }
	  else if((t4 = strstr(s + 2, "banner=off")) != NULL){
#ifdef DBG
	    fprintf(stderr, "s + 2 = %s, should be banner=off\n", s + 2);
#endif
	    banner = f;
	  }
	  else if((t4 = strstr(s + 2, "banner")) != NULL){
#ifdef DBG
            fprintf(stderr, "s + 2 = %s, just banner\n", s + 2);
#endif
	    banner = t;
	    end = LAST;
	  }
	  break;
      case 'Z':
	if((t4 = strstr(s + 2, "transparency")) != NULL){
#ifdef DBG
	  fprintf(stderr, "s + 2 = %s, should be transparency\n", s + 2);
#endif
	  trans = t;
	}
	else if((t4 = strstr(s + 2, "glossy")) != NULL){
#ifdef DBG
	  fprintf(stderr, "s + 2 = %s, should be glossy\n", s + 2);
#endif
	  glossy = t;
	}
          break;
        default:
	  break;
      }
    }
  }

  /*  Process Environment variables for banner page file */
  while(env[j] != NULL)
  {
#ifdef DBG
    fprintf(stderr, "env[%d] = %s\n", j, env[j]);
#endif
    /*  PRINTCAP_ENTRY holds all Printcap settings for a printer */
    if((t1 = strstr(env[j], "PRINTCAP_ENTRY")) != NULL)
    {
#ifdef DBG
    fprintf(stderr, "found PRINTCAP_ENTRY in env[%d]\n", j);
#endif
      if(((t2 = strstr(env[j], ":bp")) == NULL) && (banner))
      {
        fprintf(stderr, "Banner page requested, but no banner page specified in printcap, aborting!\n");
        return (-1);
      }
      else if((t2 == NULL) && (!banner));
      else if((t2 != NULL) && (banner))
      {
#ifdef DBG
    fprintf(stderr, "here is t2 after call to strstr for :bp: %s\n", t2);
#endif
        if((t3 = strstr(t2, "\n")) == NULL)
        {
	  fprintf(stderr, "Error finding end of filename for bannerpage\n");
          return(-1);
        }
#ifdef DBG
    fprintf(stderr, "here is t3 after call to strstr for \n: %s\n", t3);
    fprintf(stderr, "t2 + 4: %s\n", t2 + 4);
#endif
        bannername = (char *)calloc(t3 - (t2 + 4), sizeof(char));
        memcpy(bannername, t2 + 4, t3 - (t2 + 4));
        break;
      }
    }
    j++;
  }
      
#ifdef DBG
  if(banner)
  {
    fprintf(stderr, "in begin: bannername = %s\n", bannername);
  }
#endif      

  
/*  set the done and abrt arrays to all false */
  for(i = 0; i < 3; i++){
    done[i] = f;
    abrt[i] = f;
  }

/*  set the buffer control structure to all false and empty */
  for(i = 0; i < CHUNK + 1; i++){
    ctl[i].wash = f;
    ctl[i].size = 0;
  }
  
#ifdef DBG
  fprintf(stderr, "calling gethostbyname\n");
#endif

  while(1){
    if((prent = gethostbyname(pname)) == NULL){
#ifdef DBG
  fprintf(stderr, "h_errno = %d\n", h_errno);
#endif

      switch(h_errno)
	{
	case HOST_NOT_FOUND:
	  fprintf(stderr, "Error: Printer hostname not found, hostname: %s\n", pname);
          return (-1);
          break;
	case TRY_AGAIN:
	  break;
	case NO_RECOVERY:
	  fprintf(stderr, "Error:  Unable to recover from gethostbyname call, hostname: %s\n", pname);
          return (-1);
          break;
	case NO_DATA:
	  fprintf(stderr, "Error: No valid DNS data available for host: %s\n", pname);
          return (-1);
          break;
        default:
          fprintf(stderr, "Error: Unrecognized error generated in gethostbyname, retrying, hostname: %s\n", pname);
	  break;
	}
    }
    else break;
  }


#ifdef DBG
  fprintf(stderr, "after gethostbyname call\n");
#endif

/*  Implement more error checking code here */
/*  set up a TCP socket address */
  memcpy(&ipaddr.s_addr, *prent->h_addr_list, sizeof(ipaddr.s_addr));
  bzero(&praddr, sizeof(praddr));
  praddr.sin_family = htons(AF_INET);
  praddr.sin_port = htons(TCPPORT);
  praddr.sin_addr.s_addr = htonl(inet_addr(inet_ntoa(ipaddr)));

/*  set up a UDP socket address */
  bzero(&udppraddr, sizeof(udppraddr));
  udppraddr.sin_family = htons(AF_INET);
  udppraddr.sin_port = htons(UDPPORT);
  udppraddr.sin_addr.s_addr = praddr.sin_addr.s_addr;  

/*  zero the pagecount variables */
  p1 = p2 = 0;

/*  Needs error checking code here because it is in no way guaranteed 
 *  that calloc can get enough memory with PJOBSZ
 */
  pjob = (char *)calloc(PJOBSZ, sizeof(char));

  return 0;
}

/*  pagereq - main pagerequest thread function.  This function
 *  calls the pagerequest routine a, gets a time stamp, 
 *  calls the pagerequest routine a, gets a second time stamp,
 *  and terminates.  I have had trouble with moveing localtime_r
 *  into this thread in conjuction with the powercycle of the 
 *  printer.  The accounting(logging etc..) should appear
 *  in this function.  Until I resolve that issue, it will remain
 *  in main().   It would be useful for each thread to have valid
 *  exit status instead of pthread_exit(NULL).  Tack that on to the
 *  more error checking list.
 */
void *pagereq(void *in)
{
  int ares, diff;
  time_t rt;
  struct tm *atime;
  char *elist[] =
  {
    "Error: cannot get pagecount request in thread pagreq\n",
    "Error: unable to open accounting file",
    "Error: cannot connect to printer for second pagecount request\n",
    "Error: cannot connect to printer for first pagecount request\n"
  };


#ifdef DBG
  fprintf(stderr, "pagecount thread running\n");
#endif  


/*  call the pagerequest function a 
 *  loop until have pagerequest
 */
  pthread_mutex_lock(&preq);
  while(1){
/*  ares == -2 indicates that a() could not connect to
 *   the printer to get a page count
 */
    if((ares = a(&p1)) == -2){
      fprintf(stderr, "%s", elist[3]);
      sleep(1);
    }
    else if(ares < 0){
      fprintf(stderr, "%s", elist[0]);
      abrt[0] = t;
      pthread_exit(NULL);
    }
    else break;
  } 
  pthread_mutex_unlock(&preq);

#ifdef DBG
  fprintf(stderr, "pagecount thread past first pagecount request\n");
  fprintf(stderr, "pagecount thread calling timestamp\n");
#endif  
  
  gettimeofday(&t1, NULL); 

  sleep(1);

#ifdef DBG
  fprintf(stderr, "pagecount thread past first timestamp\n");
#endif

/*  make the second pagerequest function call a 
 *  then loop if the printer has been power cycled
 *  until it comes back up
 */  
  while(1){
/*  ares == -2 indicates that a() could not connect to
 *   the printer to get a page count
 */
    if((ares = a(&p2)) == -2){
      fprintf(stderr, "%s", elist[2]);
      sleep(1);
    }
    else if(ares < 0){
      fprintf(stderr, "%s", elist[0]);
      abrt[0] = t;
      pthread_exit(NULL);
    }
    else break;
  } 
#ifdef DBG
  fprintf(stderr, "pagecount thread past second pagecount request\n"); 

  fprintf(stderr, "pagecount thread running second timestamp\n");
#endif

    gettimeofday(&t2, NULL); 

#ifdef DBG 
  fprintf(stderr, "pagecount thread past second timestamp\n");
#endif

/* mark the thread finished */
  done[0] = t;
#ifdef DBG
  fprintf(stderr, "pagecount thread terminating\n");
#endif
  pthread_exit(NULL);
}

/*  a - this function makes the actual pagerequest from the printer.
 *  a opens a socket, connects to the printer, sends the pagerequest
 *  postscript, listens for the reply, translates the reply, closes the
 *  socket down.
 */  
int a(int *pin)
{
  int readn, writen, sres, i, j;
  size_t nleft;
  char *ptr, *tmp, *start, *page1;
  
  char *elist[] = 
  {
    "Socket Error:  Cannot get socket descriptor in thread pagereq\n",
    "Connect Error: Cannot connect to socket error in thread pagereq\n",
    "Write Error: Cannot write to socket in thread pagereq\n",
    "Read Error: Cannot read socket in thread pagereq\n"
    /* */
  };
/*  grab the socket mutex to prevent write thread from trying to write
 *  while the pagerequest is occuring
 */ 
  pthread_mutex_lock(&skt);

#ifdef DBG  
  fprintf(stderr, "STARTING PAGECOUNT REQUEST\n");
  fprintf(stderr, "In thread paqereq, sleeping 1, to clear socket\n");
#endif
  sleep(1);
/*  create the socket */
  if((prsocket = socket(AF_INET, SOCK_STREAM, 0)) < 0){
    fprintf(stderr, "%s", elist[0]);
    abrt[0] = t;
    close(prsocket);
    pthread_mutex_unlock(&skt);
    return -1;
  }

#ifdef DBG
  fprintf(stderr, "PAST SOCKET OPEN\n");
#endif
/*  connect to the printer */
  if((sres = connect(prsocket, (struct sockaddr *) &praddr, sizeof(praddr))) < 0){
#ifdef DBG
    fprintf(stderr, "In thread pagereq, connect failed\n");
#endif
    fprintf(stderr, "%s", elist[1]);
    abrt[0] = t;
    close(prsocket);
    pthread_mutex_unlock(&skt);
    return -2;
  }

#ifdef DBG
  fprintf(stderr, "PAST SOCKET CONNECT\n");
#endif  
  nleft = strlen(PAGECOUNT);
  ptr = (char *)calloc(nleft + 1, sizeof(char)); 
  /*  ptr = (char *)calloc(nleft + 2, sizeof(char)); */
  memcpy(ptr, PAGECOUNT, nleft);  
  tmp = ptr;
  /* tmp += nleft + 1;
  *tmp = (int)0x04;
  tmp = ptr; */
#ifdef DBG
  fprintf(stderr, "CONVERTING TO NETWORK BYTE ORDER\n");
#endif

/*  convert to network byte order */
  for(i = 0; i < nleft; i++){
    *tmp = htons(*tmp);
    tmp++;
  }
  tmp = NULL;
  free(ptr);
#ifdef DBG
  fprintf(stderr, "CONVERTED TO NETWORK BYTE ORDER\n");
#endif

/*  write the postscript page request to the printer */
  while(nleft > 0){
    if((writen = write(prsocket, ptr, nleft)) < 0){
      perror("write error while writing postscript pagerequest to printer");
      fprintf(stderr, "%s", elist[2]);
    /*  abrt[0] = t; */ /* changed 12-8-98, see below */
      close(prsocket);
      pthread_mutex_unlock(&skt);
/*      return -1; */
      return -2; /* changed 12-8-98 to recover from odd occurence of printer
                    lockup. To restore simply uncomment the return -1; line
		    as well as the abrt[0] = t; line */
    }
    nleft -= writen;
    ptr += writen;
  }
#ifdef DBG
  fprintf(stderr, "PAST WRITE OF REQUEST TO PAGECOUNT\n");
#endif
/*  prepare for the printer's response */
  ptr = (char *)calloc(30, sizeof(char));
  if(ptr == NULL){
    if(errno == ENOMEM)fprintf(stderr, "errno = ENOMEM\n");
    else if(errno == EAGAIN)fprintf(stderr, "errno = EAGAIN\n");
  }

  tmp = ptr; 
  nleft = 29;
#ifdef DBG
  fprintf(stderr, "nleft = %d\n", nleft);
#endif
/*  receive the printer's response */
  while(nleft > 0){
    if((readn = read(prsocket, ptr, nleft)) < 0){
      perror("read error while reading printer response from pagecount request");
      fprintf(stderr, "%s", elist[3]);
     /* abrt[0] = t; */ /* changed 12-8-98, see below */
      fprintf(stderr, "tmp = %s\n", tmp);
      close(prsocket);
      pthread_mutex_unlock(&skt);
     /* return -1; */
      return -2;  /* changed 12-8-98 to fix wierd behavior during a printer
		     hang.  This should force the filter to check for the 
		     pagecount until someone fixes the printer. To restore
 		     simply uncomment the return -1; line and the abrt[0] =t;
                     line */
    }
    else if(readn == 0) break;
    nleft -= readn;
    ptr += readn;
  }
#ifdef DBG
  fprintf(stderr, "IN A, CHANGING FROM NETWORK TO HOST BYTE ORDER\n");
#endif
  ptr = tmp;
/*  convert to host byte order from network byte order */
  for(i = 0; i < 30; i++){
    *ptr = ntohs(*ptr);
    ptr += i;
  }

#ifdef DBG
  fprintf(stderr, "tmp = %s\n", tmp);
  fprintf(stderr, "PAST LISTEN FOR PAGECOUNT RETURN\n");
#endif
 

/*  check the return string for the first numeral,
 *  this will be the start of the pagecount
 */

  ptr = tmp;
  j = 0;
  for(i = 0; i < 30; i++){
    if((*tmp >= '0') && (*tmp <= '9')){
      start = tmp;
      while((*tmp >= '0') && (*tmp <= '9')){
	j++;
        tmp++;
      }
      page1 = (char *)calloc(j, sizeof(char));
      memcpy(page1, start, j);
      break;
    }
    tmp++;
  }
#ifdef DBG
  fprintf(stderr, "AFTER PROCESSING OF PAGE NUMBER\n");
  fprintf(stderr, "page1 = %s\n", page1); 
#endif
/*  close the socket gracefully */
  shutdown(prsocket, SHUT_WR);
  while(readn != 0){
    readn = read(prsocket, ptr, 30);
  }
#ifdef DBG
  fprintf(stderr, "SOCKET SHUTDOWN COMPLETE\n");
#endif
  *pin = atoi(page1);

/*  socket is closed so we may release the socket mutex. */
  pthread_mutex_unlock(&skt);
#ifdef DBG
  fprintf(stderr, "pin = %d\n", *pin);
#endif 
  tmp = NULL;
  free(ptr);
  free(page1);
#ifdef DBG
  fprintf(stderr, "EXITING A\n");
#endif
  return 0;
}

/*  readinp - this function defines the read thread algorithm.
 *  the function detaches the thread and then starts the read loop.
 *  th read loop consists of the following steps; 1. check the
 *  buffer control structure for the current bucket to see if the
 *  bucket can have data read into it. If so, read the data from
 *  stdin into the bucket, modify the buffer control structure,
 *  and increment the bucket counter.  If not, loop until that
 *  bucket is cleaned and can have data written to it.  At this
 *  point if looping, check to see if the write thread has 
 *  terminated unnaturally(use abrt). IF so, then shut down the
 *  read thread.
 */ 
void *readinp(void *in)
{
  int i, j, readn;
  bl w;
  char *tmp;
  
  char *elist[] =
  { 
    "Error:  Unable to open data file: ",
    "Error:  Error in read of data file in thread readinp\n",
    "Error:  Error in read of data file, unnatural abort of write thread\n"
  };

  pthread_detach(pthread_self());

  i = 0;
  while(1){

#ifdef DBG
    fprintf(stderr, "In thread readinp, looping, i = %d\n", i);
#endif
/*  check the buffer control structure */
    pthread_mutex_lock(&wsh);

#ifdef DBG
    fprintf(stderr, "In thread readinp, I've got mutex wsh\n");
#endif

    if(ctl[i].wash == f){
       w = f;

#ifdef DBG
       fprintf(stderr, "ctl[%d].wash = f\n", i);
#endif

    }
    else{
       if((done[2] == t) || (abrt[2] == t)) break;
       w = t;
#ifdef DBG
       fprintf(stderr, "ctl[%d].wash = t\n", i);
#endif

    }
    pthread_mutex_unlock(&wsh);
    if(!w){
#ifdef DBG
      fprintf(stderr, "In thread readinp, I'm not waiting\n");
#endif

/*  read in the bucket of data */
      if((readn = read(0, pjob + (i * IOBUFFSZ), IOBUFFSZ)) < 0){
	fprintf(stderr, "%s", elist[1]);
	abrt[1] = t;
	pthread_exit(NULL);
      }
#ifdef DBG
      fprintf(stderr, "In thread readinp, readn = %d\n", readn);
#endif

/*  check to see if that is the end of the data */
      if(readn == 0){
        break;
#ifdef DBG
        fprintf(stderr, "In thread readinp, breaking\n");
#endif
      }
/*  change the data to network byte order in the bucket */
      tmp = pjob + (i * IOBUFFSZ);
      for(j = 0; j < readn; j++){
	*tmp = htons(*tmp);
	tmp++;
      }

/*  mark the bucket as washable */
      pthread_mutex_lock(&wsh);
      ctl[i].wash = t;
      ctl[i].size = readn;
      pthread_mutex_unlock(&wsh);
      if(i == CHUNK) i = 0;
      else i++;
    }
  else sleep(1);
  }

/*  if the write thread aborted unnaturally, abort unnaturally */
  if(abrt[2] == t){
    fprintf(stderr, "%s", elist[2]);
    abrt[1] = t;
    pthread_exit(NULL);
  }

  done[1] = t;
  pthread_exit(NULL);
}

/*  writeop - executes the algorithm that defines the write thread 
 *  writeop performs the following steps: 1. create a socket 2. connect
 *  to the printer 3. then mirror the read thread's loop and behavior
 *  only write data from buckets instead of into buckets. Also marks
 *  the buckets as free to use again.
 */
void *writeop(void *in)
{
  int writen, i, sres, nleft;
  int on = 1, j;
  bl w;
  char *tmp, *t2;
  char *elist[] =
  {
    "Error: Unable to open socket in thread writeop\n",
    "Error: Unable to connect to socket in thread writeop\n",
    "Error: Unable to write data to printer in thread writeop. Errno = ",
    "Error: Read thread abrted prematurely, unable to finish write, closing down\n",
    "Error: Unable to setsockopt SO_KEEPALIVE in thread writeop\n"
  };

  pthread_detach(pthread_self());
#ifdef DBG  
  fprintf(stderr, "In thread writeop, Starting\n");
  fprintf(stderr, "In thread writeop, getting skt\n");
#endif

/*  prevent this thread from writing to the printer until the inital page
 *  count is obtained in pagerequest thread.
 */
  pthread_mutex_lock(&preq);

/*  prevent the pagerequest thread from grabbing the socket while
 *  sending the job to the printer.
 */
  pthread_mutex_lock(&skt);

  pthread_mutex_unlock(&preq);
#ifdef DBG
  fprintf(stderr, "In thread writeop, got skt\n");
  fprintf(stderr, "In thread writeop, Sleeping to clear socket\n");
#endif

  sleep(1);

#ifdef DBG
  fprintf(stderr, "In thread writeop, I have the mutex skt\n");
#endif  
/*  create the socket */
  if((prsocket = socket(AF_INET, SOCK_STREAM, 0)) < 0){
    fprintf(stderr, "%s", elist[0]);
    abrt[2] = t;
    pthread_mutex_unlock(&skt);
    pthread_exit(NULL);
  }

/*  set the so_keepalive socket option */
  if(setsockopt(prsocket, SOL_SOCKET, SO_KEEPALIVE, (void *) &on, sizeof(on)) < 0){
    fprintf(stderr, "%s\n", elist[4]);
    abrt[2] = t;
    close(prsocket);
    pthread_mutex_unlock(&skt);
    pthread_exit(NULL); 
  }

#ifdef DBG
  fprintf(stderr, "In thread writeop, past socket call\n");
#endif

/*  connect to the socket */
  if((sres = connect(prsocket, (struct sockaddr *) &praddr, sizeof(praddr))) < 0){

#ifdef DBG
    fprintf(stderr, "In thread writeop, connect call failed\n");
#endif

    fprintf(stderr, "%s", elist[1]);
    abrt[2] = t;
    close(prsocket); 
    pthread_mutex_unlock(&skt);
    pthread_exit(NULL);
  } 

#ifdef DBG
  fprintf(stderr, "In thread writeop, past connect call\n");
#endif

/*  check for banner, print if necessary */
  if(banner){
    if(end == FIRST){
      if(genbanner(prsocket, BANNERTRAY) < 0){
        fprintf(stderr, "Error: Unable to generate banner page\n");
        abrt[2] = t;
        close(prsocket);
        pthread_mutex_unlock(&skt);
        pthread_exit(NULL);
      }
    }
  }
    
/*  check for transparency, change to transparency tray */
  if(trans){
    if(tselect(prsocket, TRANSTRAY) < 0){
      fprintf(stderr, "Error:  Unable to complete tray select call for Transparency tray\n");
      abrt[2] = t;
      close(prsocket);
      pthread_mutex_unlock(&skt);
      pthread_exit(NULL);
    }
  }

/*  check for glossy paper, change to glossy tray */
  if(glossy){
    if(tselect(prsocket, GLOSSYTRAY) < 0){
      fprintf(stderr, "Error:  Unable to complete tray select call for Glossy tray\n");
      abrt[2] = t;
      close(prsocket);
      pthread_mutex_unlock(&skt);
      pthread_exit(NULL);
    }
  }


/*  start the write iterations */
  i = 0;
  while(1){

#ifdef DBG
    fprintf(stderr, "In thread writeop, looping, i = %d\n", i);
#endif

/*  check the buffer control structure for access to the current bucket */
    pthread_mutex_lock(&wsh);
    if(ctl[i].wash == t){
      w = f;

#ifdef DBG
      fprintf(stderr, "In thread writeop, ctl[%d].wash == t, not waiting\n", i);
#endif

    }
    else{
      if((done[1] == t) || (abrt[1] == t)){
#ifdef DBG
        fprintf(stderr, "In thread writeop, read thread finished, breaking\n");
#endif
        break;
      }
      w = t;
   }
    pthread_mutex_unlock(&wsh);

    if(!w){
/*  write a bucket to the socket */
      tmp = pjob + (i * IOBUFFSZ);
      nleft = ctl[i].size;
      while(nleft > 0){
        if((writen = write(prsocket, tmp, nleft)) < 0){
          fprintf(stderr, "%s%d\n", elist[2], errno);
#ifdef DBG
          fprintf(stderr, "In thread writeop, Calling perror\n");
#endif
          perror("In thread writeop, perror says");
#ifdef DBG
          fprintf(stderr, "In thread writeop, Back from perror\n");
#endif
          abrt[2] = t;
          close(prsocket);
	  pthread_mutex_unlock(&skt);
          pthread_exit(NULL);
        }
#ifdef DBG
        fprintf(stderr, "In thread writeop, wrote %d\n", writen);
#endif
        nleft -= writen;
	tmp += writen;
      }
/*  mark the current bucket as clean */
      pthread_mutex_lock(&wsh);
      ctl[i].wash = f;
      ctl[i].size = 0;
      bzero(pjob + (i *IOBUFFSZ), IOBUFFSZ);
      pthread_mutex_unlock(&wsh);
      if(i == CHUNK)i = 0;
      else i++;
    }
  else sleep(1);
  }

/*  include more error checking here */
  tmp = malloc(RBUFFSZ);
#ifdef DBG
  fprintf(stderr, "In thread writeop, starting socket shutdown\n");
#endif

/*  check for banner, if need to print last, print it */
  if(banner){
    if(end == LAST){
     if(genbanner(prsocket, BANNERTRAY) < 0){
      fprintf(stderr, "Error: Unable to generate banner page\n");
      abrt[2] = t;
      close(prsocket);
      pthread_mutex_unlock(&skt);
      pthread_exit(NULL);
     }
    }
  }

/*  check to see if printed on transparency or glossy, if so, reset printer to default
 *  paper trays */
  if(trans || glossy){
    if(tselect(prsocket, ANYTRAY) < 0){
      fprintf(stderr, "Error, Unable to reset printer to default tray usage\n");
      abrt[2] = t;
      close(prsocket);
      pthread_mutex_unlock(&skt);
      pthread_exit(NULL);
    }
  }

/*  close socket down gracefully.  At this point, one could
 *  add code to write the statement returned from the printer
 *  to a file, or check them for errors, etc...  I have chosen
 *  to ignore them for the time being.  Hopefully a later 
 *  version will do something more useful with them.
 *
 *  Changed 11-24-98, strings returned from printer are written
 *  to log file by default
 */ 
  shutdown(prsocket, SHUT_WR);
  while(writen != 0){
    writen = read(prsocket, tmp, RBUFFSZ);
    t2 = tmp;
    for(j = 0; j < writen; j++){
      *t2 = ntohs(*t2);
      t2++;
    }
    fprintf(stderr, "Returned from Printer: %s\n", tmp);
    bzero(tmp, RBUFFSZ);
  }
#ifdef DBG
  fprintf(stderr, "In thread writeop, socket shutdown complete\n");
#endif
  free(tmp);
/*  socket is closed so release the socket mutex */
  pthread_mutex_unlock(&skt);
  if(abrt[1] == t){
#ifdef DBG
    fprintf(stderr, "In thread writeop, aborting unnaturally, due to abort in read\n");
#endif
    fprintf(stderr, "%s", elist[3]);
    abrt[2] = t;
    pthread_exit(NULL);
  }
  else{
#ifdef DBG
    fprintf(stderr, "In thread writeop, terminating naturally\n");
#endif
    done[2] = t;
    pthread_exit(NULL);
  }
}

/*  cls frees up dyanmically allocated global variables. */
int cls()
{
  free(jfilename);
  free(acctfilename);
  free(pname);
  free(usname);
  free(hname);
  free(pjob);

}

/*  CON
 *    Con takes an integer input and a pointer to a string.  It returns the string 
 *    representation of the integer.  It also assumes that the the integer will
 *    always be <= 59 because we are only using this for time related checking.
 */
char *con(int in, char *rtrn)
{
  int tmp;

  if(rtrn == NULL) return NULL;
  if(in <= 9){
    *rtrn = '0';
    *(rtrn + 1) = 0x30 + in;
  } 
  else if((in > 9) && (in <= 59)){
    tmp = in / 10;
    *rtrn = 0x30 + tmp;
    tmp = in - (tmp * 10);
    *(rtrn + 1) = 0x30 + tmp;
  }
  else{
      rtrn = "99";
  }

  return rtrn;
}


/*  ACCOUNT
 *    Account does the following items: opens the accounting file, calculates the
 *    pages printed, writes two log strings, closes the accounting file.  Account
 *    returns -1 if the accounting file cannot be opened, 0 otherwise.
 */
int account(void)
{
  int i, diff;
  struct tm atime;
  FILE *afile;
  char *buff[4];
  char *elist[] = 
  {
     "Error: unable to open accounting file"
  };


/*  open the accounting file */
  if((afile = fopen(acctfilename, "a")) == NULL){
    fprintf(stderr, "%s,%d\n", elist[0], errno);
    return -1;
  } 


#ifdef DBG
  fprintf(stderr, "altzone = %u; timezone = %u; daylight = %u\n", altzone, timezone, daylight);
  fprintf(stderr, "Errno = %d\n", errno);
#endif 


/*  calculate the number of pages printed */
  diff = p2 - p1;
  if(banner){
    if(diff == 0);
    else diff--;
  }

#ifdef DBG
  fprintf(stderr, "Calling localtime\n");
#endif 

  localtime_r((time_t *)&t1.tv_sec, &atime);

#ifdef DBG
  fprintf(stderr, "Returned from first localtime call\n");
  fprintf(stderr, "Writing first accounting string\n");
#endif

  for(i = 0; i < 4; i++){
    buff[i] = (char *)calloc(4, sizeof(char));
  }

/*  write first entry to accounting file */
  fprintf(afile, "start -p'%d' -q'%d' -k'%s' -n'%s' -h'%s' -P'%s' -F'o' -t'%s %s %s:%s:%s'\n", p1, jnumber, jfilename, usname, hname, pname, mnth[atime.tm_mon], con(atime.tm_mday,buff[0]), con(atime.tm_hour,buff[1]), con(atime.tm_min,buff[2]), con(atime.tm_sec,buff[3]));
  
#ifdef DBG
  fprintf(stderr, "Returned from first account writing\n");
  fprintf(stderr, "Calling second localtime\n");
#endif
  
  localtime_r((time_t *)&t2.tv_sec, &atime);

#ifdef DBG
  fprintf(stderr, "Returned from second localtime call\n");
  fprintf(stderr, "Writing second accounting string\n");
#endif

  for(i = 0; i < 4; i++){
    bzero(buff[i],4);
  }

/*  write second entry to accounting file */
  fprintf(afile, "end -b'%d' -p'%d' -q'%d' -k'%s' -n'%s' -h'%s' -P'%s' -F'o' -t'%s %s %s:%s:%s'\n", diff, p2, jnumber, jfilename, usname, hname, pname, mnth[atime.tm_mon], con(atime.tm_mday, buff[0]), con(atime.tm_hour,buff[1]), con(atime.tm_min,buff[2]), con(atime.tm_sec,buff[3]));
 
#ifdef DBG
  fprintf(stderr, "Returned from writing second accounting string\n");
#endif

  for(i = 0; i < 4; i++)free(buff[i]);

  return 0;
}


/*  PSTATUS
 *    pstatus opens a udp socket to the printer and checks to see if the printer
 *    is at status idle.  If so, pstatus returns 0, if an error occurs pstatus 
 *    returns -1. pstatus will wait timeout seconds for the printer to respond
 *    to the status request.
 */
int pstatus(int timeout)
{
  int udpsd, i, len;
  char *rbuff, *tmp;
  struct timeval t;
  fd_set desc;

/*  need more error checking */
  rbuff = (char *)calloc(RBUFFSZ, sizeof(char));
  len = sizeof(udppraddr);

  t.tv_sec = timeout;
  t.tv_usec = 0;

/*  create UDP socket */
  if((udpsd = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
    fprintf(stderr, "Error: Unable to create socket to check printer's status\n");
    free(rbuff);
    return (-1);
  }

/*  set up select controls */
  FD_ZERO(&desc);
  FD_SET(udpsd, &desc);
  
/*  send initial probe for printer status */
  if(sendto(udpsd, STATUS, sizeof(STATUS), 0, (struct sockaddr *) &udppraddr, len) < 0){
    fprintf(stderr, "Error: Unable to send status check command to printer\n");
    return (-1);
  }
 

/*  wait for response from printer */
  if((i = select(udpsd + 1, &desc, NULL, NULL, &t)) < 0){
    fprintf(stderr, "Error: Unable to do timeout wait for status idle\n");
    free(rbuff);
    return(-1);
  }
  else if(i == 0){
    fprintf(stderr, "Error: Unable to get printer status, request timed out\n");
    free(rbuff);
    return(-1);
  }
  
/*  get response from printer */
  if((i = recvfrom(udpsd, rbuff, RBUFFSZ, 0, (struct sockaddr *) &udppraddr, &len)) < 0){
    fprintf(stderr, "Error: Error getting response from printer\n");
    return(-1);
  }
  
#ifdef DBG
  fprintf(stderr, "rbuff = %s\n", rbuff);
#endif


/*  interpret printer status */
  if((tmp = strstr(rbuff, "status: idle")) == NULL){
    fprintf(stderr, "Error: Printer not at status idle\n");
    fprintf(stderr, "  Printer reporting the following: %s\n", rbuff);
    free(rbuff);
    return (-1);
  }
  
  free(rbuff);
  return 0;
}


/*  tselect
 *    A call to tselect sets the printer to use a specified paper tray as
 *    default.  Valid tray options are found in tray.h.
 */
int tselect(int sd, int t)
{
  char *lineout, **tray, *tmp;
  int inarray, i, j, wrote, nleft;

#ifdef DBG
  fprintf(stderr, "Starting tselect()\n");
#endif
  lineout = (char *)calloc(MLINE, sizeof(char));
  switch(t)
    {
    case AT:
      tray = anytray;
      inarray = ATS;
#ifdef DBG
      fprintf(stderr, "Switching to any tray\n");
#endif
      break;
    case LT:
      tray = lowrtray;
      inarray = LTS;
#ifdef DBG
      fprintf(stderr, "Switching to lower tray\n");
#endif
      break;
    case MT:
      tray = middtray;
      inarray = MTS;
#ifdef DBG
      fprintf(stderr, "Switching to middle tray\n");
#endif DBG
      break;
    case UT:
      tray = upprtray;
      inarray = UTS;
#ifdef DBG
      fprintf(stderr, "Switching to upper tray\n");
#endif
      break;
    case TT:
      tray = trantray;
      inarray = TTS;
#ifdef DBG
      fprintf(stderr, "Switching to Transparency tray\n");
#endif  
      break;
    case PT:
      tray = papertray;
      inarray = PTS;
#ifdef DBG
      fprintf(stderr, "Switching to Paper tray\n");
#endif
      break;
    default:
      tray = anytray;
      inarray = ATS;
#ifdef DBG
      fprintf(stderr, "As Default, Switching to any tray\n");
#endif
      break;
    }

/*  send postscript tray change code to the printer */
  for(i = 0; i < inarray; i++){
#ifdef DBG
    fprintf(stderr, "In main for loop, iteration: %d\n", i);
    fprintf(stderr, "Calling memcpy\n");
#endif
    memcpy(lineout, tray[i], (nleft = strlen(tray[i])));
#ifdef DBG
    fprintf(stderr, "Past memcpy\n");
    fprintf(stderr, "lineout = %s\n", lineout);
    fprintf(stderr, "nleft = %d\n", nleft);
#endif DBG
    tmp = lineout;
#ifdef DBG
    fprintf(stderr, "Starting conversion to network byte order\n");
#endif DBG

/*  convert to network byte order */
    for(j = 0; j < MLINE; j++){
      *tmp = htons(*tmp);
      tmp++;
    }

#ifdef DBG
    fprintf(stderr, "Past conversion to network byte order\n");
#endif
    tmp = lineout;
#ifdef DBG
    fprintf(stderr, "Starting write of string to printer\n");
#endif

/*  write data to printer */
    while(nleft > 0){
      if((wrote = write(sd, lineout, nleft)) < 0){
	fprintf(stderr, "Error: In trayselect, unable to write to socket\n");
        free(lineout);
        return (-1);
      }
      tmp += wrote;
      nleft -= wrote;
    }

#ifdef DBG
    fprintf(stderr, "Past write of string to printer\n");
    fprintf(stderr, "Calling bzero\n");
#endif
    bzero(lineout, MLINE);
#ifdef DBG
    fprintf(stderr, "Past bzero call\n");
#endif
  }

#ifdef DBG
  fprintf(stderr, "Returning from tselect()\n");
#endif
  free(lineout);
  return (0);
}
  
/*  genbanner
 *    This function generates  a banner page on the specified tray.  Tray options are defined
 *    in tray.h. 
 *    The following tags should appear in the banner page; **PNAME**, **TYPE**, **UNAME**, **HNAME**,
 *    **DT**.  These tags correspond to the printer name, printer type, user name, host name,
 *    and date/time respectively. genbanner reads a file specified in bannername, substitutes
 *    job specific information for the tags and sends the banner page to the printer.
 *    After printing the banner page, genbanner will switch the printer to use the default paper tray.
 */
int genbanner(int sd, int tray)
{

  int bannerfile, banbufsize = 16384, readn, wrote, nleft, i, j, ts;
  unsigned long retref;
  char *bannerbuf, *t1, *t2;

  char *vname[] =
  {
    "**PNAME**",
    "**TYPE**",
    "**UNAME**",
    "**HNAME**",
    "**JNAME**",
    "**DT**"
  };
  

  if(bannername == NULL){
    fprintf(stderr, "Error: banner page requested but no filename found.\n");
    return(-1);
  }

  /*  open bannerfile */
  if((bannerfile = open(bannername, 0)) < 0)
  {
    perror("banner file open");
    return(-1);
  }

  if((ts = tselect(sd, tray)) < 0){
    fprintf(stderr, "Error: unable to select the appropriate tray for banner page\n");
    return(-1);
  }

  /*  read bannerfile, check for tags, write data to printer */
  while(1)
  {

    /*  bannerbuf gets modified in routine insrt, to guarantee a read of at most
        banbufsize, we free bannerbuf at the end of the loop and reinitialize it at
	the beginning. */
    bannerbuf = (char *)calloc(banbufsize, sizeof(char));

    if((readn = read(bannerfile, bannerbuf, banbufsize)) < 0)
    {
      perror("Reading banner file");
      return(-1);
    }
    /*  check for end of file from bannerfile */
    if(readn == 0)
    {
      free(bannerbuf);
      break;
    }

    /*  look for tags in bannerbuf */
    for(j = 0; j < 6; j++)
    {
      while(1)
      {

        if((t1 = strstr(bannerbuf, vname[j])) != NULL)
	{
	  /*  insrt replaces the tag with print specific information 
	      The bannerbuf = (char *)retref insures that bannerbuf 
	      points to the newly created string from insrt  */
	  insrt(bannerbuf, t1, j, &readn, &retref); 
          bannerbuf = (char *)retref;

	}
        else break;
      }
    }
    /*  make sure bannerbuf is null terminated */
    strcat(bannerbuf, "\0");
    readn++;
#ifdef DBG
  fprintf(stderr, "Past insrt calls, readn = %d\nbannerbuf = %s\n", readn,bannerbuf);
#endif
    t2 = bannerbuf;
#ifdef DBG
  fprintf(stderr, "now t2 = %s\n", t2);
#endif
  /*  convert bannerbuf to network byte order */
    for(j = 0; j < readn; j++)
    {
     
#ifdef DBG
      /*  fprintf(stderr, "j = %d, t2[%d] = %c\n", j, j, *t2); */
#endif
      *t2 = htons(*t2);
      t2++;
    }
#ifdef DBG
  fprintf(stderr, "Past change to network byte order\n");
#endif

  /* write bannerbuf to the printer */
    nleft = readn;
    t1 = bannerbuf;
    while(nleft > 0)
    {
      if((wrote = write(sd, t1, nleft)) < 0)
      {
        perror("writing banner to printer");
        free(bannerbuf);
        return(-1);
      }
      t1 += wrote;
      nleft -= wrote;
    }
    free(bannerbuf);
    bannerbuf = (char *)calloc(banbufsize, sizeof(char));
  }
   
#ifdef DBG
  fprintf(stderr, "Past write of banner to socket\n");
#endif

  if((ts = tselect(sd, ANYTRAY)) < 0){
    fprintf(stderr, "Error: Unable to reset default papertray after banner page\n");
    return(-1);
  }

  return(0);

}


/*  insrt 
 *    This function takes string line, a pointer to the tag in line, and
 *    replaces the tag with the appropriate useful job information.  IE,
 *    for **PNAME** insrt substitutes pname- the printer hostname etc...
 *    insrt leaves line with the tag removed and the appropriate info inserted.
 *    insrt destroys the orinal line string and allocates a new string. The
 *    variable ret holds the pointer address of the new string in memory to
 *    insure that the calling routine can sync up.  Also, inln holds the number
 *    of bytes in the new string for the calling routine to use.  
 */
void insrt(char *line, char *rep, int sw, int *inln, unsigned long *ret)
{
  char *temp, *rstring;
  int i, length, out;
  time_t now;
  bl rs = f;

#ifdef DBG
  fprintf(stderr, "inln: %d, sw = %d\n", *inln, sw);
#endif


/*  determine which tag we have */
  switch(sw)
    {
    case 0:
      rstring = pname; 
      out = 9;
#ifdef DBG
      fprintf(stderr, "insrt: found **PNAME**\n");
#endif
      break;
    case 1:
      rstring = "Tektronix Phaser560";
      out = 8;
#ifdef DBG
      fprintf(stderr, "insrt: found **TYPE**\n");
#endif
      break;
    case 2:
      rstring = usname;
      out = 9;
#ifdef DBG
      fprintf(stderr, "insrt: found **UNAME**\n");
#endif
      break;
    case 3:
      rstring = hname;
      out = 9;
#ifdef DBG
      fprintf(stderr, "insrt: found **HNAME**\n");
#endif
      break;
    case 4:
      rstring = jfilename;
      out = 9;
#ifdef DBG
      fprintf(stderr, "insrt: found **JNAME**\n");
#endif
      break;
    case 5:
      rstring = (char *)calloc(30, sizeof(char));
      now = time(NULL);
      ctime_r(&now, rstring, 30);
      out = 6;
      rs = t;
#ifdef DBG
      fprintf(stderr, "insrt: found **DT**\n");
#endif
      break;
    default:
      break;
    }

#ifdef DBG
  fprintf(stderr, "insrt: rstring = %s\n", rstring);
  fprintf(stderr, "insrt: out = %d\n", out);
#endif

  /*  calculate the length of the string after removing the tag */
  length = *inln - out; 


#ifdef DBG
  fprintf(stderr, "insrt: length = %d\n", length);
#endif

  /*  allocate memory for the new string with print job information */
  temp = (char *)calloc(length + strlen(rstring), sizeof(char));

  
#ifdef DBG
  fprintf(stderr, "insrt: rep - line = %d\n", rep - line);
#endif

  /*  copy info from old string to the new string up to the point of the tag */
  memcpy(temp, line, rep - line);


#ifdef DBG
  /*  fprintf(stderr, "insrt: after memcpy, temp = %s\n", temp); */
#endif

  /*  copy the print specific information into the new string */
  memcpy(temp + (rep - line), rstring, strlen(rstring));

#ifdef DBG
   fprintf(stderr, "insrt: strlen(rstring) = %d\n", strlen(rstring));
   fprintf(stderr, "insrt: starting memcopy for leftover bytes\n");
   fprintf(stderr, "insrt: rep + out = %s\n", rep + out);
   fprintf(stderr, "insrt: (line + inln) - (rep + out) + 1 = %d\n", (line + *inln) - (rep + out));
#endif

   /*  copy the info from the old string to the new string that occurs after the tag */
  memcpy(temp + (rep - line) + strlen(rstring), rep + out, (line + *inln) - (rep + out)); 


#ifdef DBG
  fprintf(stderr, "insrt: after memcopy of leftover bytes\n");
  fprintf(stderr, "insrt: leftover bytes copied = %d\n", (line + *inln) - (rep + out));
#endif


  /*  Release the old string's memory, allocate new memory, update variable information,
      copy the locally formed new string into the new space.  update pointer information,
      free up local copy of new string */
  free(line);
  line = (char *)calloc(length + strlen(rstring), sizeof(char));
  *inln = length + strlen(rstring);
  memcpy(line, temp, length + strlen(rstring));
  *ret = (unsigned long)line;

  if(rs) free(rstring);
  free(temp);
}
